/*
 *   senator - ASF video stream server
 *   Copyright (C) 2002  anonymous author
 *
 *
 *  senator is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *   
 *  Senator is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *	
 *  You should have received a copy of the GNU General Public License
 *  along with senator; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 *
 */
    
#include <stdlib.h>
#include <stdio.h>
#include <list>

#include "CStreamHandler.h"
#include "global.h"
#include "netclass/ncmutex.h"

extern	char *itoa(int i, char *a, int r);
extern	list <void *>	connectionList;
extern	ncMutex	updateConnectionList;

char *ipToStr(unsigned long ip){
    struct in_addr inAddr;
    memset(&inAddr,0,sizeof(inAddr));
    inAddr.s_addr = ip;
    return inet_ntoa(inAddr);
} 


CStreamHandler::CStreamHandler(ncSocket *io){
	this->_client		= io;
	this->_sequence		= 0;
	this->_ID		= -1;
	this->_currentPacket	= 0;
	this->_avarageDelay	= 0;
	_asfFile		= NULL;
	_clientVersion 		= "unknown";
	_playing		= false;
}

CStreamHandler::~CStreamHandler(){
#ifdef STREAM_HANDLER_DEBUG_OUTPUT    
    printf("deleting stream handler\n");
#endif 

    if (_asfFile)
        delete _asfFile;
    if (_client)
	_client->close();
}

/*
 * handleRequest
 *
 * reads the request from the client
 *
 */
void CStreamHandler::handleRequest(){
	memset(buf,0,NUMBYTES);
	int numActualBytes = 0;
	
	_clientIP = ipToStr(_client->getIpAddr());

	logger->logInfo("["+_clientIP+"] client connected");
	
	// read client requenst
	switch(_client->readData(buf,NUMBYTES-1,&numActualBytes)){
		case NC_OK:
		case NC_MOREDATA:
			if (numActualBytes < 100) return;
#ifdef STREAM_HANDLER_DEBUG_OUTPUT
			printf("----------------------------------------\n");
			printf("%s\n",buf);
			printf("----------------------------------------\n");
			printf("parsing request\n");
#endif		    
			parseHTTPRequest(buf);

#ifdef STREAM_HANDLER_DEBUG_OUTPUT
			printf("----------------------------------------\n");
			printf("done\n");
#endif
			break;
		default:
			logger->logText("["+_clientIP+"] read from client failed");
			break;
	}
}

/*
 *  parse a HTTP request
 *
 *  this function can handle the GET command and check
 *  if the client want to start the stream, check for
 *  a file or skip in the stream
 *
 *  The stream-time is the timecode (milliseconds) for seeking within the stream.
 *  I don't know if this is correct but the request goes like this:
 *
 *  if the stream-time is 0:0 the client expects the ASF file header
 *  
 *    Pragma: no-cache,rate=1.000000,stream-time=0,stream-offset=0:0,request-context=1,max-duration=0
 *
 *  otherwise, it request an time in the file
 *
 *    Pragma: no-cache,rate=1.000000,stream-time=0,stream-offset=0:768172,request-context=2,max-duration=217376
 *
 *  if the clients first sends a ?, it tries to connect using the TCP protocol...
 *  this is not supported by the server so we have to wait for the next connection,
 *  this will be a HTTP request.. 
 */
void CStreamHandler::parseHTTPRequest(const char *buf){
	char	cmd[1024];
	char	file[1200];
	char	ver[1024];
        int	num;

	mplayerStatus	state;
	state = requestStreamData;		// default

	if (connectionList.size() > (unsigned int)config->maxConnections){
	    logger->logInfo("to many connections, please edit the config file");
	    sendHTTPError(serverFull);
	    return;
	}

	num = sscanf(buf, "%s %s %s", cmd, file, ver);

	// check the command
	if (strncmp(cmd,"GET",3) == 0){
	    if (config->statusPageName.length() > 0 && strlen(file) == config->statusPageName.length())
		if (strncmp(file,config->statusPageName.c_str(),config->statusPageName.length()) == 0){
			logger->logInfo("["+_clientIP+"] sending client status page");
			_fileName = config->statusPageName;
			sendStatusPage();
			return;
		}
		this->_asfFile = new CASFFile((std::string)file);

		switch(_asfFile->open()){
		    case ILLEGAL_REQUEST:
			logger->logInfo("["+_clientIP+"] client made an illegal request (403): "+(std::string)file);
			sendHTTPError(accessForbidden); // we are being to nice :)
			return;
			break;
		    case FILE_NOT_FOUND:
			logger->logInfo("["+_clientIP+"] file not found (404): "+(std::string)file);
			sendHTTPError(fileNotFound);
			return;
			break;
		}

		_asfFile->parseFile();

		_clientVersion = getPragmaValue(buf,"User-Agent");
		std::string requestContext = getPragmaValue(buf,"request-context");
		if (atoi(requestContext.c_str()) == 1)
			state = requestFileHeader;

		switch(state){
			case requestFileHeader:
				logger->logInfo("["+_clientIP+"] client requested fileheader for : " + (std::string)file);
				sendHTTPResponse();
				sendFileHeader();
				break;
			case requestStreamData:
			
				_fileName = (string) file;
				
				string streamTime = getPragmaValue(buf, "stream-time");
				if (streamTime.length() == 0)
					streamTime = "0";
#ifdef STREAM_HANDLER_DEBUG_OUTPUT
				printf ("streamtime = %s\n",streamTime.c_str());
#endif
				logger->logInfo("["+_clientIP+"] starting stream of : " + (std::string)file + " at "+streamTime+"ms");

				sendHTTPResponse();
				sendFileHeader();
				sendStreamData(atoi(streamTime.c_str()));
				break;
		}
	} else {
#ifdef STREAM_HANDLER_DEBUG_OUTPUT
		printf("not a GET command, ignoring\n");
#endif
	}
}

/*
 * getPragamValue
 *
 * parse a value out of the HTTP header
 */
std::string CStreamHandler::getPragmaValue(const char *buf,const std::string &key){

	char *pragmaOffset = 0;
	// scan for the pragma code
	pragmaOffset = strstr(buf,key.c_str());
	if ( pragmaOffset == 0){
		return "";
	}

	// get the value
	pragmaOffset+=key.length()+1;

	std::string result;
	int i=0;
	while(pragmaOffset[i] != (char)',' && pragmaOffset[i] != (char)'\r' && pragmaOffset[i] != (char)'\n'){
	    result += pragmaOffset[i];
	    i++;
	}

#ifdef STREAM_HANDLER_DEBUG_OUTPUT
	printf("pragma %s, value %s\n",key.c_str(),result.c_str());
#endif
	    
	return result;
}


/*
 *  sendHTTPError
 *
 *  return an error to the client
 *
 */
void CStreamHandler::sendHTTPError(const HTTPErrors &e){
	std::string errorText;
	char	string[10];
	
	itoa(e,string,10);
	
	std::string errorNr = string;

	errorText ="HTTP/1.0 "+ errorNr +"\r\n";
	errorText+="Server: "+ config->identifyAs +"\r\n";
	errorText+="Content-type: text/html\r\n";
	errorText+="\r\n";

	this->_client->writeData((char*)errorText.c_str(),errorText.length());
}

/*
 *  sendHTTPResponse
 *
 *  file found and we can read it
 *
 */
void CStreamHandler::sendHTTPResponse(){
	std::string response;
	logger->logText("["+_clientIP+"] sending client the HTTP 200 header");

	response = "HTTP/1.0 200 OK\r\n";
	response+= "Content-Type: application/octet-stream\r\n";
	response+= "Server: "+config->identifyAs+"\r\n";
	response+= "Cache-Control: no-cache\r\n";
	response+= "Pragma: no-cache\r\n";
	response+= "Pragma: client-id=132854\r\n";
	response+= "Pragma: features=\"seekable,stridable\"\r\n";
	response+= "Connection: close\r\n";
	response+= "\r\n";

	this->_client->writeData((void*)response.c_str(),response.length());
}

/*
 *  generateStatusPage
 *
 *  show some server stats
 *
 */
void CStreamHandler::sendStatusPage(){

	updateConnectionList.lock();		// prevent altering of the connection list
						// while we are reading it
	char	clientConnectionCount[200];
	itoa(connectionList.size(),clientConnectionCount,10);
	
	char	maxClients[20];
	itoa(config->maxConnections,maxClients,10);
	char	serverPort[20];
	itoa(config->portNumber,serverPort,10);

	std::string r = "";
	r+= "<html><head><title>senator "+ (std::string)VERSION + " status page</title></head>";
	r+= "<body bgcolor=\"#000000\" text=\"#ffffff\" link=\"#ffffff\" vlink=\"#ffffff\">";
	r+= "<font face=\"Verdana\" size=\"5\" color=\"#FFCC00\">";
	r+= "senator " + (std::string)VERSION +"</font>";
	r+= "<br><font size=1><a href=\"http://www.angelfire.com/linux/senator/\">latest version</a></font><hr>";

	r+= "<font face=\"Verdana\" size=\"1\" color=\"#ffffff\">";
	r+= "<table width=500>";
	r+= "<tr><td>server identifies</td><td>" + config->identifyAs +"</td></tr>";
	r+= "<tr><td>server port</td><td>" + (string) serverPort +"</td></tr>";
	r+= "<tr><td>clients connected</td><td>"+ (string) clientConnectionCount +"</td></tr>";
	r+= "<tr><td>max connections</td><td>"+ (string) maxClients +"</td></tr>";
	r+= "<tr><td>root directory</td><td>"+ config->rootDirectory +"</td></tr>";
	r+= "<tr><td>protocol type</td><td>";

	if (config->HTTPStreamProtocol)
	    r+="HTTP stream protocol";
	else r+="raw";
	r+=" </td></tr>";
	
	r+= "</table>";
	r+= "</font><br><br>";

	r+= "<font face=\"Verdana\" size=\"3\" color=\"#FFCC33\">";
	r+= "client enumeration<br>";
	r+= "</font>";
	
	if (connectionList.size() > (unsigned int)0) {
	
		r+= "<font face=\"Verdana\" size=\"1\" color=\"#ffffff\">";
		r+= "<center><table width=570 border=0>";

		r+= "<tr bgcolor=\"#cccccc\">";
		r+= "<td width=70>nr</td>";
		r+= "<td width=100>client type</td>";
		r+= "<td width=100>client IP</td>";
		r+= "<td with=200>playing</td>";
		r+= "<td width=100>percent</td>";
		r+= "</tr>";

		float value;
		int precision = 2;
		int decimal, sign;

		list<void *>::const_iterator iter;
		unsigned int i = 0;
		for (iter=connectionList.begin(),i = 0; i < connectionList.size(); i++,iter++) {
			CStreamHandler *walker = (CStreamHandler*)(*iter);
			itoa(i+1,clientConnectionCount,10);
			
			r+= "<tr>";
			r+= "<td>" + (string)clientConnectionCount + "</td>";
			r+= "<td>" + walker->getClientVersion() + "</td>";
			r+= "<td>" + walker->getClientIP() + "</td>";
			r+= "<td>" + walker->getPlayFileName() + "</td>";
			
			// percent
			value = (float)walker->getPlayPercent();

			if (value > 0){
    			    string percent = fcvt(value, precision, &decimal, &sign);
			    if (decimal > 0)	// quick patch
				r+= "<td>" + percent.substr(0,decimal)+",";
			    else
				r+= "<td>0,";
			    r+= percent.substr(decimal,percent.length()) + "%</td>";
			} else {
				r+= "<td>n/a</td>";
			}
			r+= "</tr>";
		}

		r+="</table></center>";
		r+= "</font>";
	}
	r+= "</body></html>";

	updateConnectionList.unlock();			// release the lock

	sendHTTPResponse();
	this->_client->writeData((char*)r.c_str(),r.length());
}


/*
 *  sendFileHeader
 *
 *  sends the file header from the requested asf file
 *
 */
void CStreamHandler::sendFileHeader(){
	unsigned short headerSize = _asfFile->getASFHeaderSize();
	unsigned char *buffer = new unsigned char [headerSize];
	_asfFile->getASFHeader(buffer);

	sendProtocolPacketHeader(streamHeader,headerSize,0x0c00);
	this->_client->writeData(buffer,headerSize);
}

/*
 *  sendFileData
 *
 *  sends the actual stream data and provides scrolling
 *
 *  connectionLost is used to check if there is still a 
 *  connection with the player. MediaPlayer disconnects
 *  after the first request. This could be detected in
 *  a better way I guess.
 *
 */
 
#define DELAY_CORRECTION 15
 
void CStreamHandler::sendStreamData(unsigned long offset){
	
	_playing = true;		// check if we need to send % data for status page
	
	FILEHEADER	fileHeader;
	
	this->_asfFile->getFileHeader(&fileHeader);

	unsigned char *dataBuffer = new unsigned char [fileHeader.packetSize];
	
	int actualBytes = 0;

	// The stream-time is the timecode (milliseconds) for seeking within the stream
	// calculate the packet number with the correct stream-time

	ASF_uint64 temp		= fileHeader.playDuration / 10000;
	ASF_uint64 temp2	= fileHeader.numPackets;
	double  packetDelta	= (double)(signed)temp2 / (double)(signed)temp;
	int   	packetNumber	= 0; 

	if (offset > 0)
		packetNumber = (int) (packetDelta * (double) offset);
	 else
		packetNumber = 0;
	
	_sequence = 0;

#ifdef STREAM_HANDLER_DEBUG_OUTPUT
	printf("timeoffset   = %d (ms)\n",(unsigned int)offset);
	printf("packetNumber = %d (start) / %d (total)\n",packetNumber,(unsigned int)fileHeader.numPackets);
#endif
	unsigned char	*pp		= NULL;
	char		pktflags 	= 0;
	unsigned short	padsize 	= 0;
	unsigned short	frame_duration	= 0;
	unsigned long	packet_send_time= 0;
	int		segments_expected = 0;

	_currentPacket = packetNumber;
	
	_client->flush();

	bool is_key = false;

	for (int i=packetNumber;i<fileHeader.numPackets;i++){
		this->_asfFile->getFrame(dataBuffer,i);
		
		// find out the frame size for this packet
		pp = (unsigned char *)dataBuffer;
		padsize = 0;

		if (*(short *)pp != (short)'YS') {
	    	    pktflags = pp[3];
		    pp+=5;
		    if (pktflags & 0x10) {	// 0x10: padding flag
			padsize += *(unsigned short *)pp;
			pp+=2;
		    }

    		    if (pktflags & 0x08)
	    		padsize += *pp++;
		
		    packet_send_time = *(unsigned long*)pp;
		    pp+=4;
		    frame_duration = *(unsigned short *)pp;
		    pp+=2;
    		    if (pktflags & 0x01) {
			    segments_expected = *pp++ & 0x7f;
		    }
							    
		    is_key = !!(pp[0] & 0x80); 
		    
		} else {
		    printf("chubby packet found!\n");
		}
		
		is_key = !!(pp[0] & 0x80);
		_avarageDelay+=frame_duration;
		_avarageDelay =_avarageDelay >> 1;

		actualBytes = 0;	// used check if the connection is still there

		
#ifdef STREAM_HANDLER_TALK_HAPPY
		printf("segments expected = %d\n",segments_expected);
		printf("packet send: %d (%d of total %d) - sleeping %d ms - avarage delay %d\n",this->_sequence,i,(unsigned int)fileHeader.numPackets,(unsigned short)frame_duration - DELAY_CORRECTION,_avarageDelay);
		printf("packet send time %d",packet_send_time);
#endif		

		// send the actual data
		sendProtocolPacketHeader(streamData,fileHeader.packetSize,0x0000);
		this->_client->writeData(dataBuffer,fileHeader.packetSize,&actualBytes);

		// check if there is still a connection
		if (fileHeader.packetSize != (unsigned int)actualBytes){
#ifdef STREAM_HANDLER_DEBUG_OUTPUT
    		    printf("\n\n\nconnection with client lost\n");
		    printf("needed to send %d / could send %d\n",fileHeader.packetSize,actualBytes);
#endif
		    _playing = false;
		    return;
		}

		// stream delay
		if (frame_duration - DELAY_CORRECTION > 0 && _sequence > 10)
		    ncSleep((unsigned short) (frame_duration - DELAY_CORRECTION ));
		    
		this->_sequence++;
		this->_currentPacket++;
	}

	sendProtocolPacketHeader(streamEnd,0,0x0000);
	_playing = false;
}

/*
 * low level protocol function
 *
 * type 	2 bytes
 * length	2 bytes
 * sequence	4 bytes
 * res.		2 bytes (0)
 * length	2 bytes
 * (data	-)
 *
 */
void CStreamHandler::sendProtocolPacketHeader(const packetType &t,const unsigned short &length,const unsigned short &part){	
	unsigned short res  = 0;
	unsigned short type = t;
	
	if (config->HTTPStreamProtocol == true){

	    this->_client->writeData(&type,1);		// packet type
	    type = type>>8;
	    this->_client->writeData(&type,1);

	    unsigned short tl = length + 8;
	    this->_client->writeData(&tl,1);		// packet length
	    tl = tl >> 8;
	    this->_client->writeData(&tl,1);

	    unsigned int ts = _sequence;
	    this->_client->writeData(&ts,1);		// sequence number
	    ts = ts >> 8;
	    this->_client->writeData(&ts,1);
	    ts = ts >> 8;
	    this->_client->writeData(&ts,1);
	    ts = ts >> 8;
	    this->_client->writeData(&ts,1);

	    res = part;					// what is this?
	    this->_client->writeData(&res,1);
	    res = res >> 8;
	    this->_client->writeData(&res,1);

	    tl = length + 8;
	    this->_client->writeData(&tl,1);		// packet length
	    tl = tl >> 8;
	    this->_client->writeData(&tl,1);
	}
}

/*
 *  class interface routines
 *
 *
 */
string CStreamHandler::getPlayFileName(){
	return this->_fileName;
}

void CStreamHandler::setConnectionID(const int &ID){
	this->_ID = ID;
}

int CStreamHandler::getID(){
	return _ID;
}

double CStreamHandler::getPlayPercent(){
	FILEHEADER	fileHeader;
	
	if (this->_playing == true){
    	    this->_asfFile->getFileHeader(&fileHeader);

	    double total	= (signed)fileHeader.numPackets;
	    double current	= (signed)_currentPacket;
	    double percent	= (current / total) * 100.0f;
	    return percent;
	    

	} 
	return -1;
}

double CStreamHandler::getKBPerSec(){
	FILEHEADER	fileHeader;
	this->_asfFile->getFileHeader(&fileHeader);
	
	double size = fileHeader.packetSize;
	return (1000.0f / (double)_avarageDelay) * size;
}

std::string CStreamHandler::getClientVersion(){
    return _clientVersion;
}

std::string CStreamHandler::getClientIP(){
    return _clientIP;
}